/* chil2nc.c -- converts Chilbolton-format radar data to netCDF
   A multitude of processing options are available

   Copyright (C) 2000 Robin Hogan
*/

#include <netcdf.h>
#include "chil.h"

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    CHILPrintf("NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); \
    exit(-1); \
}

#define MIN(a, b) ((a) < (b) ? (a) : (b))

void
add_attributes(int ncid, int fieldid, char *units, char *long_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
}

int
main(int argc, char **argv)
{
  int status, iarg = 0;
  ChilboltonData data;
  ChilboltonFile *chil_file = &(data.chil_file);
  RasterInfo *raster_info = &(data.raster_info);
  Ray *ray = &(data.ray);
  char aname[20];
  char cloud_data, output_nez = 0;
  char *tmp_name, *id = "";

  int rayid, gateid, freqid, altid, latid, lonid;
  int ncid, rangeid, elevid, azimid, nezid, timeid, fieldid[MAX_FIELDS];
  size_t start[2], count[2];
  size_t tmp_index;
  int dimids[2];
  int index = 0;
  unsigned short tmp_uword;
  float tmp_float;
  unsigned short startgate, endgate, nfields = 1;

  status = CHILInit(argc, argv, &data);

  if (argc < 2)
    {
      CHILUsage(&data);
      CHILOptions(&data);
      exit(0);
    }

  /* Look for netCDF-related arguments */
  
  while (++iarg < argc) {
    if (strcmp(argv[iarg],"-id") == 0) {
      if (argc > iarg+1) {
	id = argv[++iarg];
      }
      else {
	CHILPrintf("Error: id not specified\n");
	exit(1);
      }
    }
  }

  if (status == CHIL_OK)
    {

      while (1)
	{
	  int i, j;
	  char *version = CHIL_VERSION;
	  int failures = 0, maxfailures = 10;

	  while ((status = CHILFetchRay(&data)) != CHIL_OK)
	    {
	      if (status != CHIL_ENDOFBLOCK) {
		break;
	      }
	      if ((++failures) == maxfailures)
		{
		  CHILPrintf("Aborting: %d failed calls to CHILFetchRay - empty scans?\n");
		  break;
		}
	    }
	  if (status != CHIL_OK)
	    {
	      break;
	    }
	    

	  cloud_data = raster_info->cloud_data;
	  if (cloud_data)
	    {
	      sprintf(aname,"%04hu%02hu%02hu_%d%s.nc",
		      CHILTrueYear(ray->year),
		      ray->month, ray->day, (int) raster_info->frequency,
		      id);
	      if (data.process) {
		output_nez = 1;
	      }
	    }
	  else
	    {
	      sprintf(aname,"%04hur%03hus%02hu%s.nc", raster_info->file,
		      raster_info->raster, chil_file->iscan+1, id);
	    }

	  /* Open netCDF file for writing */
	  handle_error(nc_create(aname, NC_CLOBBER, &ncid));
	  index = 0;
	  if (!data.quiet)
	    CHILPrintf("Writing %s\n", aname);

	  /* Add some global attributes */
	  tmp_name = CHILSystemName(raster_info, NULL);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system",
				       strlen(tmp_name), tmp_name));
  
	  tmp_name = CHILScanType(raster_info, NULL);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "scantype",
				       strlen(tmp_name), tmp_name));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT,
					1, &(ray->day)));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT,
					1, &(ray->month)));

	  tmp_uword = CHILTrueYear(ray->year);
	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT,
					1, &tmp_uword));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "file",
					NC_SHORT,
					1, &(raster_info->file)));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "raster",
					NC_SHORT,
					1, &(raster_info->raster)));
 	  if (!cloud_data) {
	    tmp_uword = chil_file->iscan + 1;
	    handle_error(nc_put_att_short(ncid, NC_GLOBAL, "scan",
					  NC_SHORT,
					  1, &tmp_uword));
	  }
	  tmp_name = CHILGetOptions(&data);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "options",
				       strlen(tmp_name), tmp_name));
	  CHILFree(tmp_name);

	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "software_version",
				       strlen(version), version));

	  /* Add time/ray dimension */
	  handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &rayid));

	  /* Add range/gate dimension  */
	  endgate = ray->endgate;
	  nfields = MIN(ray->ntotal, data.maxfields);
	  startgate = ray->startgate;
	  handle_error(nc_def_dim(ncid, "range", endgate - startgate, &gateid));

	  /* Add some scalar variables */
	  handle_error(nc_def_var(ncid, "frequency", NC_FLOAT, 0, dimids, 
				  &freqid));
	  add_attributes(ncid, freqid, "GHz", "Radar frequency");

	  handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
				  &latid));
	  add_attributes(ncid, latid, "degrees_north", NULL);

	  handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
				  &lonid));
	  add_attributes(ncid, lonid, "degrees_east", NULL);

	  handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, dimids, 
				  &altid));
	  add_attributes(ncid, altid, "m",
			 "Altitude of antenna above mean sea level");

	  /* Add time coordinate variable */ 
	  dimids[0] = rayid;
	  handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
				  &timeid));
	  add_attributes(ncid, timeid, "hours UTC", "Decimal hours since midnight");

	  /* Add range coordinate variable */
	  dimids[0] = gateid;
	  handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1, dimids, &rangeid));
	  add_attributes(ncid, rangeid, "km", "Range from antenna");

	  /* If not cloud data, add elevation and azimuth */
	  dimids[0] = rayid;
	  if (!cloud_data)
	    {
      	      handle_error(nc_def_var(ncid, "elev", NC_FLOAT, 1, dimids,
				    &elevid));
	      add_attributes(ncid, elevid, "degrees",
			    "Elevation above horizon");

	      handle_error(nc_def_var(ncid, "azim", NC_FLOAT, 1, dimids,
				    &azimid));
	      add_attributes(ncid, azimid, "degrees",
			     "Azimuth clockwise from due north");
	    }
	  else if (output_nez)
	    {
      	      handle_error(nc_def_var(ncid, "nez", NC_FLOAT, 1, dimids,
				    &nezid));
	      add_attributes(ncid, nezid, "dBZ",
			    "Noise-equivalent reflectivity factor at 1 km");	      
	    }
	  dimids[0] = rayid;
	  dimids[1] = gateid;

	  tmp_float = chil_no_signal_value;
	  for (i = 0; i < nfields; i++)
	    {
	      handle_error(nc_def_var(ncid, CHILFieldName(ray, i, NULL),
				  NC_FLOAT, 2, dimids, &fieldid[i]));
	      add_attributes(ncid, fieldid[i], CHILFieldUnits(ray, i, NULL),
			     CHILFieldDescription(ray, i, NULL));
	      handle_error(nc_put_att_float(ncid, fieldid[i], "missing_value",
					    NC_FLOAT, 1, &tmp_float));

	    }

	  /* End define mode */
	  handle_error(nc_enddef(ncid));

	  /* Set the scalar variables */
	  tmp_index = 0;
	  handle_error(nc_put_var1_float(ncid, freqid, &tmp_index,
					 &(raster_info->frequency)));

	  tmp_float = CHILAltitude(raster_info);
	  handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &tmp_float));

	  tmp_float = CHILLatitude(raster_info);
	  handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &tmp_float));

	  tmp_float = CHILLongitude(raster_info);		       
	  handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &tmp_float));

	  /* Set the range */
	  for (j = startgate; j < endgate; j++)
	    {
	      tmp_index = j - startgate;
	      tmp_float = ray->range[j] + 0.001 * ray->range_offset;
	      handle_error(nc_put_var1_float(ncid, rangeid,
					     &tmp_index, &tmp_float));
	    }

	  do
	    {
	      tmp_index = index;
	      handle_error(nc_put_var1_float(ncid, timeid, &tmp_index,
					   &(ray->utc)));
	      if (!cloud_data)
		{
		  handle_error(nc_put_var1_float(ncid, elevid, &tmp_index,
					       &(ray->elevation)));
		  /* Make azimuth from due north rather than due west */
		  tmp_float = ray->azimuth - 90.0;
		  handle_error(nc_put_var1_float(ncid, azimid, &tmp_index,
					       &tmp_float));
		}
	      else if (output_nez)
		{
		  handle_error(nc_put_var1_float(ncid, nezid, &tmp_index,
					       &(ray->nez)));
		}

	      for (i = 0; i < nfields; i++)
		{
		  start[0] = index;
		  start[1] = 0;
		  count[0] = 1;
		  count[1] = endgate - startgate;
		  handle_error(nc_put_vara_float(ncid, fieldid[i], start,
						 count, ray->data[i] 
						 + startgate));
		}

	      ++index;
	    }
	  while ((status = CHILFetchRay(&data)) == CHIL_OK);

	  handle_error(nc_close(ncid));

	  if (status == CHIL_ENDOFDATA)
	    {
	      break;
	    }
	}
      CHILEnd(&data);
    }
  else {
    CHILPrintf("Run with no options for more information\n");
    exit(1);
  }
  exit(0);
}
